(*
  bits - a formal system demonstrating cyclic values
  Copyright (C) 1996  Scott Draves <spot@cs.cmu.edu>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundationeither version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this programif not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

datatype Val = val_int of int
             | val_bool of bool

datatype Var = var of string

datatype Prim = p_plus | p_times | p_equal
              | p_div  | p_mod
              | p_bit_and | p_bit_or | p_bit_xor | p_bit_not
              | p_bit_shift_left | p_bit_shift_right


datatype Exp = exp_val of Val
             | exp_var of Var
             | exp_abs of Var * Exp
             | exp_fix of Var * Exp
             | exp_letrec of (Var * Exp) list * Exp * Memos
             | exp_app of Exp * Exp
             | exp_if  of Exp * Exp * Exp
             | exp_prim of Prim * Exp list
             | exp_lift of Exp
             | exp_debug of Exp

and      M = m_static of Val
           | m_dynamic of Exp
           | m_function of (M -> M)
           | m_cyclic of Val * Exp * Val

and Meq = meq_static of Val
      |   meq_cyclic of Val * Val



withtype Env = (Var -> M)
and Memos = (*  ((Env * M) list ref) *)
                ((Meq * (Var * M)) list ref)


exception unbound of Var

exception bad_bases of Prim * int * int


fun print_val (val_int n) = (print "(val_int "; print n; print ")")
  | print_val (val_bool n) = (print "val_bool "; print n; print")")

fun print_var (var v) = (print "(var "; print v; print ")")

fun print_prim p_plus = print "p_plus"
  | print_prim p_times = print "p_times"
  | print_prim p_equal = print "p_equal"
  | print_prim p_div = print "p_div"
  | print_prim p_mod = print "p_mod"
  | print_prim p_bit_and = print "p_bit_and"
  | print_prim p_bit_or = print "p_bit_or"
  | print_prim p_bit_xor = print "p_bit_xor"
  | print_prim p_bit_not = print "p_bit_not"
  | print_prim p_bit_shift_left = print "p_bit_shift_left"
  | print_prim p_bit_shift_right = print "p_bit_shift_right"


fun print_exp_list (hd :: tail) = (print_exp hd; print_exp_list tail)
  | print_exp_list _ = ()

and print_exp (exp_val v) = (print "(exp_val "; print_val v; print ")")
  | print_exp (exp_var v) = (print "(exp_var "; print_var v; print ")")
  | print_exp (exp_abs(v,e)) = (print "(exp_abs "; print_var v; print " "; print_exp e; print ")")
  | print_exp (exp_fix(v,e)) = (print "(exp_fix "; print_var v; print " "; print_exp e; print ")")
  | print_exp (exp_app(e0,e1)) = (print "(exp_app "; print_exp e0; print " "; print_exp e1; print ")")
  | print_exp (exp_if(e0,e1,e2)) = (print "(exp_if "; print_exp e0; print " "; print_exp e1; print " "; print_exp e2; print ")")
  | print_exp (exp_lift v) = (print "(exp_lift "; print_exp v; print ")")
  | print_exp (exp_debug v) = (print "(exp_debug "; print_exp v; print ")")
  | print_exp (exp_prim(p,es)) = (print "(exp_prim "; print_prim p; print " ["; print_exp_list es;  print "])")


fun print_m (m_static v) = (print "(m_static "; (print_val v); print")")
  | print_m (m_dynamic d) = (print "(m_dynamic "; print_exp d; print")")
  | print_m (m_function d) = (print "(m_function ???)")
  | print_m (m_cyclic(b,q,r)) = (print "(m_cyclic "; print_val b; print " "; print_exp q; print " "; print_val r; print")")


fun print_meq (meq_static v) = (print "(meq_static "; (print_val v); print")")
  | print_meq (meq_cyclic(b,r)) = (print "(meq_cyclic "; print_val b; print " "; print_val r; print")")
fun empty_env v = raise unbound(v)

fun extend v c env =
   fn v' => (if (v = v') then c else (env v'))


local
  val counter = ref 0
in
  fun genvar () = (inc counter;
                   var (implode["v", (makestring (!counter))]))
end

fun reify (m_static v) = (exp_val v)
  | reify (m_dynamic v) = v
  | reify (m_function f) = let val v = genvar()
                               val _ = (print "rf "; print_var v; print "\n")
                               val e = reify (f (m_dynamic (exp_var v)))
                           in exp_abs(v, e)
                           end
  | reify (m_cyclic (b, q, r)) =
           exp_prim(p_plus, [exp_prim(p_times, [q, (exp_val b)]), (exp_val r)])


fun apply (m_function f) v = (f v)
  | apply (m_dynamic s0) (m_dynamic s1) = (m_dynamic (exp_app (s0, s1)))


fun all_dynamic args = map (*fn m_dynamic x => x
                             | x => (print "static arg to dynamic prim: ";
                                     print_m x; raise Match)*)
                            reify
                            args

local
  fun mem m ((key,value)::memos) = if m = key then SOME(value) else (mem m memos)
    | mem m _ = NONE

  fun remo k ((key,x)::memos) acc =  if (k = key)
                                    then (remo k memos acc)
                                    else (remo k memos ((key,x)::acc))
    | remo k _ acc = acc
in
  fun remove_memo k memos = (memos := remo k (!memos) [])
  fun add_memo key x memos = (memos := ((key,x) :: (remo key (!memos) [])))
  fun find_memo m (ref memos) = (mem m memos)
end

fun spec (exp_val v) env = m_static v
  | spec (exp_var v) env = (env v)
  | spec (exp_lift e) env = m_dynamic (reify (spec e env))
                                 (* always unfold? *)
  | spec (exp_app(e0, e1)) env = apply (spec e0 env) (spec e1 env)

  | spec (exp_abs(v, e)) env =
        m_function (fn v' => ((* print "binding "; print_var v;
                               print_m v'; print "\n"; *)
                              (spec e (extend v v' env))))

  | spec (exp_debug e) env = let val v = (spec e env)
                             in print "debug "; print_m v; print "\n";
                                v
                             end

  | spec (exp_fix(v, e)) env =
      let fun F (m_dynamic d) =
               let val v' = genvar()
                   val e' = reify(spec e (extend v (m_dynamic d) env))
               in m_dynamic(exp_app(exp_fix(v', e'), e))
               end
             | F m = ((* print "fixing "; print_var v; print_m m; print "\n"; *)
                       (apply (spec e (extend v (m_function F) env)) m))
      in (m_function F)
      end

  | spec (exp_letrec(blocks, body, memos)) env =
      if true (* oracle *)
      then let fun buildenv ((lab,lam) :: tl) = 
                     extend lab (m_function (F lam)) (buildenv tl)
                 | buildenv _ = env
               and F e m = let val key = case m of (m_static v) => (meq_static v)
                                                 | (m_cyclic (b, q, r)) =>  (meq_cyclic (b, r))
                                                 | _ => (print "xrl "; print_m m; raise Match)
                           in case (find_memo key memos) of
                             (SOME(name,defn)) => (print "hit\n";(m_dynamic(exp_app(exp_var(name), reify(m)))))
                            | NONE => let val name = genvar()
                                     in (add_memo key (name, m_dynamic(exp_var(var("nyet")))) memos);
                                         print "mv "; print_var name; print "\n";
                                         let val rez = spec e (buildenv blocks)
                                         in (add_memo key (name, rez) memos);
                                             (print "added "; print_meq(key); print " "; print_m(rez); print "\n");
                                            (apply rez m)
                                         end
                                     end
                           end

            (* fun reify_block(v,e) = (v,reify(spec e (buildenv blocks))) *)
               fun reify_memo(key,(name,defn)) = (name,reify(defn))
	       val body' = reify(spec body (buildenv blocks))
           in print "reifying memos\n";m_dynamic(exp_letrec((map reify_memo(!memos)),
                                   body', memos))
           end
      else
        let fun buildenv ((lab,lam) :: tl) = 
                     extend lab (m_function (F lam)) (buildenv tl)
              | buildenv _ = env
            and F e m = (apply (spec e (buildenv blocks)) m)
            (* fun reify_block(v,e) = spec e (buildenv blocks) *)
        in (spec body (buildenv blocks))
        end

  | spec (exp_if(e0, e1, e2)) env =
      (case (spec e0 env) of
          (m_static (val_bool v)) => if v 
                                     then (spec e1 env)
                                     else (spec e2 env)
        | (m_dynamic d0) => case ((spec e1 env), (spec e2 env)) of
              ((m_dynamic d1),  (m_dynamic d2)) =>
                      m_dynamic (exp_if (d0, d1, d2)))

  | spec (exp_prim (p, es)) env =
      let fun plus b q r x = 
             let val r' = (r+x) mod b
                 val q' = (r+x) div b
             in if 0=q'
                then m_cyclic((val_int b), q, (val_int (r')))
                else m_cyclic((val_int b), exp_prim(p_plus,[exp_val(val_int(q')),q]), (val_int (r')))
             end
          fun times b q r x = if (x > 0) 
                 then m_cyclic((val_int (b*x)), q, (val_int (r*x)))
                 else if (x < 0) then m_cyclic((val_int (~b*x)), q, (val_int (~r*x)))
                 else m_static(val_int(0))
      in
      (case (p, (map (fn e => (spec e env)) es)) of

         (p_times, [m_static(val_int x),
                  m_cyclic((val_int b), q, (val_int r))]) => (times b q r x)
        
       | (p_times, [m_cyclic((val_int b), q, (val_int r)),
                    m_static(val_int x)]) => (times b q r x)
           

       | (p_plus, [m_cyclic((val_int b), q, (val_int r)),
                  m_static(val_int x)]) => (plus b q r x)
                      
       | (p_plus, [m_static(val_int x),
                 m_cyclic((val_int b), q, (val_int r))]) => (plus b q r x)


       | (p_equal, [m_cyclic((val_int b0), q0, (val_int r0)),
                  m_cyclic((val_int b1), q1, (val_int r1))]) =>
         if (b0 = b1)
         then let val rr0 = r0 mod b0
                  val rr1 = r1 mod b0
                  val rq0 = r0 div b0
                  val rq1 = r1 div b0
              in if (rr0 = rr1)
                 then m_dynamic(exp_prim(p_equal,
                                if (rq0 = rq1)
                                then [q0, q1]
                                else [q0, exp_prim(p_plus, [q1, exp_val(val_int (rq0-rq1))])]))
                 else m_static(val_bool false)
              end
         else raise bad_bases(p_equal, b0, b1)

       | (p_div, [m_cyclic((val_int b0), q0, (val_int r0)),
                  m_static(val_int d)]) =>
           if 0 = (b0 mod d)
           then m_cyclic((val_int (b0 div d)), q0, (val_int (r0 div d)))
           else raise bad_bases(p_div, b0, d)

       | (p_mod, [m_cyclic((val_int b0), q0, (val_int r0)),
                  m_static(val_int d)]) =>
           if (b0 = d)
           then m_static(val_int (r0 mod d))
           else raise bad_bases(p_mod, b0, d)

       | (p_plus, [m_static (val_int x), m_static (val_int y)])
           => m_static(val_int(x + y))
       | (p_times, [m_static (val_int x), m_static (val_int y)])
           => m_static(val_int(x * y))
       | (p_equal, [m_static (val_int x), m_static (val_int y)])
           => m_static(val_bool(x = y))
       | (p_div, [m_static (val_int x), m_static (val_int y)])
           => m_static(val_int(x div y))
       | (p_mod, [m_static (val_int x), m_static (val_int y)])
           => m_static(val_int(x mod y))
       | (p_bit_or, [m_static (val_int x), m_static (val_int y)])
           => m_static(val_int(Bits.orb(x, y)))
       | (p_bit_and, [m_static (val_int x), m_static (val_int y)])
           => m_static(val_int(Bits.andb(x, y)))
       | (p_bit_shift_left, [m_static (val_int x), m_static (val_int y)])
           => m_static(val_int(Bits.lshift(x, y)))
       | (p_bit_shift_right, [m_static (val_int x), m_static (val_int y)])
           => m_static(val_int(Bits.rshift(x, y)))
       | (p_bit_xor, [m_static (val_int x), m_static (val_int y)])
           => m_static(val_int(Bits.xorb(x, y)))
       | (p_bit_not, [m_static (val_int x)])
           => m_static(val_int(Bits.notb(x)))
       

       | (p, es) => m_dynamic(exp_prim(p, (all_dynamic es))))
     end


(* assume e = fn s fn d . e'  *)

fun mix e x =
    let val e_x = (apply (spec e empty_env) (m_static x))
    in (reify (spec (reify e_x) empty_env))
    end


