functor Lex(structure Token : TOKEN) : LEX =
struct
  structure Token = Token
  open Token

(*
 * Lexical analyzers for special tokens and whitespace
 *)

(* Whitespace consists of blanks, tabs, newlines, and comments *)
fun comment n = any -- (fn "{" => comment (n+1)
                         | "}" => if n = 0 then succeed ()
                                           else comment (n-1)
                         | _ => comment n)

fun whitespace x = (literal " " || literal "\t" || literal "\n"
                             || literal "{" && comment 0 return "") x

(* The "digit" function is a lexical analyzer for numeric digits.  The
 * "makeint" function converts a string of digits into an integer token *)
fun is_digit c = "0" <= c andalso c <= "9"
fun digit x    = (any suchthat is_digit wth (fn d => ord (String.sub(d,0)) - ord (#"0"))) x

fun makeint0 ds = (foldl (fn (d,n) => 0w10*n + (Word32.fromInt d)) 0w0 ds)
fun makeint  ds = tok_int (makeint0 ds)

(* The "alphabetic" and "alphanumeric" functions perform lexical analysis
 * of characters.  "makeid" is a utility for converting a string of
 * characters into an identifier token *)
fun is_alphabetic c   =  ("A" <= c andalso c <= "Z") orelse
                         ("a" <= c andalso c <= "z")
fun is_alphanumeric c = is_alphabetic c orelse c = "_" orelse is_digit c

fun alphabetic x   = (any suchthat is_alphabetic) x
fun alphanumeric x = (any suchthat is_alphanumeric) x

fun makeid (c,cs) =
    case concat (c::cs) of
        "true"   => tok_bool(true)
      | "false"  => tok_bool(false)
      | "if"     => tok_if
      | "let"    => tok_let
      | "fun"    => tok_fun
      | "loop"   => tok_loop
      | "val"    => tok_val
      | "and"    => tok_and
      | "or"     => tok_or
      | id       => tok_name id

fun minus (x:myint) = 0w0 - x

(* "token" puts all of the above together, into a general lexical
 * analyzer for Spartan *)
fun token x = (repeat1 digit wth makeint
         || literal "~" && repeat1 digit wth (fn (_,ds) => tok_int(minus (makeint0 ds)))
         || alphabetic && (repeat alphanumeric) wth makeid
         || literal "-" && literal ">" return tok_arrow
         || literal "<" && literal "<" return tok_leftleft
         || literal ">" && literal ">" return tok_rightright
         || literal ">" && literal "=" return tok_greater_equal
         || literal "<" && literal "=" return tok_less_equal
         || literal "\\" return tok_lambda
         || literal "." return tok_dot
         || literal "," return tok_comma
         || literal "+" return tok_plus
         || literal "-" return tok_minus
         || literal "*" return tok_star
         || literal "." return tok_dot
         || literal "%" return tok_percent
         || literal "/" return tok_slash
         || literal "=" return tok_equal
         || literal ":" return tok_colon
         || literal ";" return tok_semicolon
         || literal "(" return tok_lparen
         || literal ")" return tok_rparen
         || literal "&" return tok_ampersand
         || literal "<" return tok_less
         || literal ">" return tok_greater
         || literal "|" return tok_bar
         || literal "'" return tok_quote) x

(* A simple error-reporting function *)
fun lexerror pos = print ("Lexical error at " ^ Position.makestring pos ^ "\n")

(* The "lex" function skips all whitespace, searching for the next token.
 * Errors are reported by the "lexerror" function. *)
fun lex x =
    (middle (repeat whitespace) (!! token guard lexerror) (repeat whitespace)) x

(* Finally, we use the "transform" function to produce a lexical analyzer *)
val lexer = transform lex

end
